import os
from datetime import datetime
from math import ceil
from pathlib import Path

import pandas as pd
import pytest
from _pytest.logging import LogCaptureFixture
from loguru import logger
from pandas.testing import assert_frame_equal

from vivarium_cluster_tools.psimulate.paths import OutputPaths
from vivarium_cluster_tools.psimulate.performance_logger import (
    append_child_job_data,
    append_perf_data_to_central_logs,
    generate_runner_job_data,
    transform_perf_df_for_appending,
)


@pytest.fixture
def artifact_perf_df():
    filepath = Path(__file__).parent / "data/artifact_perf_df.csv"
    df = pd.read_csv(filepath)
    index_cols = ["host", "job_number", "task_number", "draw", "seed"]
    scenario_cols = [col for col in df.columns if col.startswith("scenario")]
    df = df.set_index(index_cols + scenario_cols)
    return df


@pytest.fixture
def artifactless_perf_df():
    filepath = Path(__file__).parent / "data/artifactless_perf_df.csv"
    df = pd.read_csv(filepath)
    index_cols = ["host", "job_number", "task_number", "draw", "seed"]
    scenario_cols = [col for col in df.columns if col.startswith("scenario")]
    df = df.set_index(index_cols + scenario_cols)
    return df


@pytest.fixture
def result_directory():
    return Path("/mnt/team/simulation_science/pub/models/project_name/results/model_version/")


@pytest.fixture
def caplog(caplog: LogCaptureFixture):
    handler_id = logger.add(
        caplog.handler,
        format="{message}",
        level=0,
        filter=lambda record: record["level"].no >= caplog.handler.level,
        enqueue=False,  # Set to 'True' if your test is spawning child processes.
    )
    yield caplog
    logger.remove(handler_id)


def get_output_paths_from_output_directory(output_directory):
    model_name = "artifact"
    original_launch_time = "YYYY_MM_DD_HH_MM_SS"
    launch_time = "yyyy_mm_dd_hh_mm_ss"
    output_directory = output_directory / model_name / original_launch_time

    logging_directory = output_directory / "logs" / f"{launch_time}_runtype"
    logging_dirs = {
        "logging_root": logging_directory,
        "cluster_logging_root": logging_directory / "cluster_logs",
        "worker_logging_root": logging_directory / "worker_logs",
    }

    output_paths = OutputPaths(
        root=output_directory,
        **logging_dirs,
        worker_settings=output_directory / "settings.py",
        environment_file=output_directory / "requirements.txt",
        model_specification=output_directory / "model_specification.yaml",
        keyspace=output_directory / "keyspace.yaml",
        branches=output_directory / "branches.yaml",
        finished_sim_metadata=output_directory / "finished_sim_metadata.csv",
        results_dir=output_directory / "results",
        backup_dir=output_directory / "sim_backups",
        backup_metadata_path=output_directory / "sim_backups" / "backup_metadata.csv",
    )

    return output_paths


@pytest.mark.parametrize("df_name", ["artifact_perf_df", "artifactless_perf_df"])
def test_expected_columns(df_name, result_directory, tmp_path, request):
    perf_df = request.getfixturevalue(df_name)
    # transform df
    output_paths = get_output_paths_from_output_directory(result_directory)
    central_perf_df = transform_perf_df_for_appending(perf_df, output_paths)
    expected_columns = (
        ["host", "job_number", "task_number", "draw", "seed"]
        + perf_df.columns.tolist()
        + ["artifact_name", "scenario_parameters"]
    )
    assert list(central_perf_df.columns) == expected_columns


@pytest.mark.parametrize("df_name", ["artifact_perf_df", "artifactless_perf_df"])
def test_data_parsing(df_name, result_directory, tmp_path, request):
    perf_df = request.getfixturevalue(df_name)
    output_paths = get_output_paths_from_output_directory(result_directory)
    central_perf_df = transform_perf_df_for_appending(perf_df, output_paths)
    assert (central_perf_df["artifact_name"] == "artifact").all()

    if df_name == "artifact_perf_df":
        expected_scenario_parameters = [
            '{"scenario_parameter_one": "value_one", "scenario_input_data_artifact_path": "/path/to/artifact.hdf"}',
            '{"scenario_parameter_one": "value_two", "scenario_input_data_artifact_path": "/path/to/artifact.hdf"}',
        ] * 6
    else:
        expected_scenario_parameters = [
            '{"scenario_parameter_one": "value_one"}',
            '{"scenario_parameter_one": "value_two"}',
        ] * 6
    assert (central_perf_df["scenario_parameters"] == expected_scenario_parameters).all()

    job_number = int(central_perf_df["job_number"].unique().squeeze())
    runner_data = generate_runner_job_data(job_number, output_paths, "first_file_with_data")

    assert runner_data["project_name"].squeeze() == "project_name"
    assert runner_data["root_path"].squeeze() == Path(
        "/mnt/team/simulation_science/pub/models/project_name/results/model_version/artifact"
    )
    assert runner_data["original_run_date"].squeeze() == "YYYY_MM_DD_HH_MM_SS"
    assert runner_data["run_date"].squeeze() == "yyyy_mm_dd_hh_mm_ss"
    assert runner_data["run_type"].squeeze() == "runtype"
    assert runner_data["log_summary_file_path"].squeeze() == "first_file_with_data"
    assert (
        runner_data["original_log_file_path"].squeeze()
        == "/mnt/team/simulation_science/pub/models/project_name/results/model_version/artifact/YYYY_MM_DD_HH_MM_SS/logs/yyyy_mm_dd_hh_mm_ss_runtype/worker_logs/log_summary.csv"
    )


def test_valid_log_path(result_directory, artifact_perf_df, caplog, tmp_path, monkeypatch):
    monkeypatch.setattr(
        "vivarium_cluster_tools.psimulate.performance_logger.CENTRAL_PERFORMANCE_LOGS_DIRECTORY",
        tmp_path,
    )
    monkeypatch.setattr(
        "vivarium_cluster_tools.psimulate.performance_logger.NUM_ROWS_PER_CENTRAL_LOG_FILE", 4
    )
    # add some data to central logs directory to allow appending
    output_paths = get_output_paths_from_output_directory(result_directory)
    central_perf_df = transform_perf_df_for_appending(artifact_perf_df, output_paths)
    pd.DataFrame(columns=central_perf_df).to_csv(tmp_path / "log_summary_0000.csv")
    # test no warnings were raised
    append_perf_data_to_central_logs(artifact_perf_df, output_paths)
    assert not caplog.records


@pytest.mark.parametrize(
    "invalid_log_path",
    [
        Path("/mnt/team/simulation_science/pub/models/project_name/model_version/"),
        Path("/mnt/team/simulation_science/pub/project_and_model_version/"),
        Path("/ihme/homes/user/model_version/"),
    ],
)
def test_invalid_log_path(invalid_log_path, artifact_perf_df, caplog):
    # test we raise specific warning
    output_paths = get_output_paths_from_output_directory(invalid_log_path)
    append_perf_data_to_central_logs(artifact_perf_df, output_paths)
    assert "Skipping appending central performance logs." in caplog.text


@pytest.mark.parametrize(
    "available_rows, rows_to_append, expected_output_files, multiple_log_files_exist",
    [
        (2, 2, 2, False),
        (2, 2, 2, True),
        (2, 6, 3, False),
        (2, 6, 3, True),
        (2, 9, 3, False),
        (4, 4, 2, False),
        (3, 2, 1, False),
        (1, 3, 2, False),
    ],
)
def test_appending(
    available_rows,
    rows_to_append,
    expected_output_files,
    multiple_log_files_exist,
    artifact_perf_df,
    result_directory,
    tmp_path,
    monkeypatch,
):
    max_num_rows = 4
    monkeypatch.setattr(
        "vivarium_cluster_tools.psimulate.performance_logger.NUM_ROWS_PER_CENTRAL_LOG_FILE",
        max_num_rows,
    )
    monkeypatch.setattr(
        "vivarium_cluster_tools.psimulate.performance_logger.CENTRAL_PERFORMANCE_LOGS_DIRECTORY",
        tmp_path,
    )

    # set up tests
    # create data we want to append
    output_paths = get_output_paths_from_output_directory(result_directory)
    central_perf_df = transform_perf_df_for_appending(artifact_perf_df, output_paths)
    data_to_append = central_perf_df[:rows_to_append]
    # create most recent files
    if multiple_log_files_exist:
        central_perf_df.to_csv(tmp_path / "log_summary_0000.csv", index=False)
        most_recent_file = tmp_path / "log_summary_0001.csv"
        expected_output_files += 1
    else:
        most_recent_file = tmp_path / "log_summary_0000.csv"
    initial_data = pd.DataFrame(
        index=range(max_num_rows - available_rows), columns=central_perf_df.columns
    )
    initial_data.to_csv(most_recent_file, index=False)

    # append data and test
    first_file_with_data = append_child_job_data(data_to_append)

    assert first_file_with_data == str(most_recent_file)

    # test that all files we expect to exist are there
    absolute_output_filepaths = sorted(tmp_path.glob("*"))
    output_filenames = [filepath.stem for filepath in absolute_output_filepaths]
    expected_filenames = [
        f"log_summary_{str(i).zfill(4)}" for i in range(expected_output_files)
    ]
    assert expected_filenames == output_filenames

    # test that each of those files has the right number of rows and the expected data
    # inspect first file
    first_file = pd.read_csv(most_recent_file)
    assert_frame_equal(first_file[: len(initial_data)], initial_data, check_dtype=False)
    assert_frame_equal(
        first_file[len(initial_data) :].reset_index(drop=True),
        data_to_append[: (max_num_rows - len(initial_data))],
        check_dtype=False,
    )

    data_to_append = data_to_append[(max_num_rows - len(initial_data)) :].reset_index(
        drop=True
    )

    if multiple_log_files_exist:
        # check that existing file wasn't modified
        existing_file = pd.read_csv(tmp_path / "log_summary_0000.csv")
        assert_frame_equal(existing_file, central_perf_df, check_dtype=False)
        # remove empty file from list of files to check
        absolute_output_filepaths = absolute_output_filepaths[1:]
    # inspect remaining files
    for file in absolute_output_filepaths[1:]:
        file_data = pd.read_csv(file)
        assert_frame_equal(file_data, data_to_append[:max_num_rows], check_dtype=False)
        data_to_append = data_to_append[max_num_rows:].reset_index(drop=True)



"""
==========================
vipin Performance Counters
==========================

Structs for counting performance metrics.

"""
import json
from time import time

import psutil


class CounterSnapshot:
    def __init__(self, cpu=None, disk=None, freq=None, net=None, timestamp=None):
        self.cpu = psutil.cpu_stats() if cpu is None else cpu
        self.disk = (
            psutil.disk_io_counters(perdisk=False, nowrap=True) if disk is None else disk
        )
        self.freq = psutil.cpu_freq(percpu=False) if freq is None else freq
        self.net = psutil.net_io_counters(pernic=False, nowrap=True) if net is None else net
        self.timestamp = time() if timestamp is None else timestamp

    def to_dict(self):
        c_dict = dict()
        c_dict["cpu"] = dict(self.cpu._asdict())
        c_dict["freq"] = dict(self.freq._asdict())
        c_dict["disk"] = dict(self.disk._asdict())
        c_dict["net"] = dict(self.net._asdict())
        c_dict["time"] = self.timestamp
        return c_dict

    def __sub__(self, other):
        cpu = type(other.cpu)(
            *tuple(self.cpu[i] - other.cpu[i] for i in range(len(other.cpu)))
        )
        disk = type(other.disk)(
            *tuple(self.disk[i] - other.disk[i] for i in range(len(other.disk)))
        )
        freq = type(other.freq)(
            *tuple((self.freq[i] + other.freq[i]) / 2 for i in range(len(other.freq)))
        )
        net = type(other.net)(
            *tuple(self.net[i] - other.net[i] for i in range(len(other.net)))
        )
        timestamp = self.timestamp - other.timestamp
        return CounterSnapshot(cpu, disk, freq, net, timestamp)

    def __repr__(self) -> str:
        return json.dumps(self.to_dict())



"""
=====================
Performance Reporting
=====================

Tools for summarizing and reporting performance information.

"""

import json
import re
from pathlib import Path

import numpy as np
import pandas as pd
import requests
from loguru import logger
from pandas import json_normalize

BASE_PERF_INDEX_COLS = ["host", "job_number", "task_number", "draw", "seed"]

# The number of scenario columns beyond which we shorten the scenarios to a single string
COMPOUND_SCENARIO_COL_COUNT = 2


class PerformanceSummary:
    """A class to implement a getter for data in the workers' performance logs.

    Given a Path, a PerformanceSummary class provides a generator to get at each
    entry in the workers' performance logs. The class also provides a method
    to get all entries in a pd.DataFrame. This class is intended as a singleton
    to provide data about a single Vivarium simulation run.

    Attributes
    ----------
    log_dir
        Path of log_dir
    errors
        Number of errors encountered while parsing logs

    """

    def __init__(self, log_dir: Path):
        self.log_dir: Path = log_dir
        self.errors: int = 0

    def get_summaries(self) -> dict:
        """Generator to get all performance summary log messages in PerformanceSummary"""
        for log in [
            f for f in self.log_dir.iterdir() if self.PERF_LOG_PATTERN.fullmatch(f.name)
        ]:
            with log.open("r") as f:
                count: int = 0
                for line in f.readlines():
                    count += 1
                    try:
                        message = json.loads(line)["record"]["message"]
                    except Exception as e:
                        logger.warning(
                            f"Exception: {e}. Malformed message in {log} line {count}, skipping..."
                        )
                        self.errors += 1
                        continue
                    m = self.TELEMETRY_PATTERN.fullmatch(str(message))
                    if m:
                        yield json_normalize(json.loads(message), sep="_")

    def to_df(self) -> pd.DataFrame:
        perf_data = []
        for item in self.get_summaries():
            perf_data.append(item)
        if len(perf_data) < 1:
            return pd.DataFrame()
        perf_df = pd.concat(perf_data)

        # Convert the Unix timestamps to datetimes
        for col in [col for col in perf_df.columns if col.startswith("event_")]:
            perf_df[col] = pd.to_datetime(perf_df[col], unit="s")

        # Remove trailing "_scenario" from normalized label
        perf_df.columns = perf_df.columns.str.replace("_scenario", "", regex=False)
        return perf_df

    TELEMETRY_PATTERN = re.compile(r"^{\"host\".+\"job_number\".+}$")
    PERF_LOG_PATTERN = re.compile(r"^perf\.([0-9]+)\.([0-9]+)\.log$")

    def clean_perf_logs(self):
        """Remove all performance logs from the log_dir (after to_df has been called)"""
        for log in [
            f for f in self.log_dir.iterdir() if self.PERF_LOG_PATTERN.fullmatch(f.name)
        ]:
            log.unlink()


def set_index_scenario_cols(perf_df: pd.DataFrame) -> tuple[pd.DataFrame, list]:
    """Get the columns useful to index performance data by."""
    index_cols = BASE_PERF_INDEX_COLS
    scenario_cols = [col for col in perf_df.columns if col.startswith("scenario_")]
    index_cols.extend(scenario_cols)
    perf_df = perf_df.set_index(index_cols)
    return perf_df, scenario_cols


def add_squid_api_data(perf_df: pd.DataFrame):
    """Add Squid API data to the performance dataframe.

    Given a dataframe from PerformanceSummary.to_df, add Squid API data for the job.
    Squid API reference: https://hub.ihme.washington.edu/display/SCKB/How+to+use+Squid+API
    """
    try:
        job_numbers = perf_df["job_number"].unique()
        assert len(job_numbers) == 1
        squid_api_data = requests.get(
            f"http://squid.ihme.washington.edu/api/jobs?job_ids={job_numbers[0]}"
        ).json()
        squid_api_df = pd.DataFrame(squid_api_data["jobs"]).add_prefix("squid_api_")
        perf_df = perf_df.astype({"job_number": np.int64})
        perf_df = perf_df.merge(
            squid_api_df,
            left_on=["job_number"],
            right_on=["squid_api_job_id"],
        )
    except Exception as e:
        logger.warning(f"Squid API request failed with: {e}")
    return perf_df


def print_stat_report(perf_df: pd.DataFrame, scenario_cols: list):
    """Print some helpful stats from the performance data.

    The stats are grouped by scenario_cols.
    """
    pd.set_option("display.max_rows", None)
    pd.set_option("display.max_columns", None)
    pd.options.display.float_format = "{:.2f}".format

    do_compound = len(scenario_cols) > COMPOUND_SCENARIO_COL_COUNT

    perf_df = perf_df.reset_index()

    if do_compound:
        logger.info(
            f"compound scenario:\n({'/'.join([s.replace('scenario_', '') for s in scenario_cols])}):"
        )
        perf_df["compound_scenario"] = (
            perf_df[scenario_cols]
            .to_csv(header=None, index=False, sep="/")
            .strip("\n")
            .split("\n")
        )

    # Print execution times stats by scenario
    temp = (
        perf_df.set_index("compound_scenario" if do_compound else scenario_cols)
        .filter(like="exec_time_")
        .stack()
        .reset_index()
    )

    if do_compound:
        cols = ["compound_scenario", "measure", "value"]
    else:
        cols = scenario_cols
        cols.extend(["measure", "value"])

    temp.columns = cols
    cols.remove("value")

    report_df = temp.groupby(cols).describe()
    report_df.columns = report_df.columns.droplevel()
    report_df = report_df.drop(["count", "25%", "50%", "75%"], axis=1)
    report_df = report_df.reset_index()

    # Abbreviate execution time measures for printing
    report_df["measure"] = report_df["measure"].replace("^exec_time_", "", regex=True)
    report_df["measure"] = report_df["measure"].replace(
        "^simulant_initialization", "sim_init", regex=True
    )
    report_df["measure"] = report_df["measure"].replace("minutes$", "min", regex=True)
    report_df["measure"] = report_df["measure"].replace("seconds", "s", regex=True)

    report_df = report_df.set_index(cols).sort_index()
    logger.info(f"\n{report_df}")


def report_performance(
    input_directory: Path | str,
    output_directory: Path | str,
    output_hdf: bool,
    verbose: int,
):
    """Main method for vipin reporting.

    Gets job performance data, outputs to a file, and logs a report.
    """
    input_directory, output_directory = Path(input_directory), Path(output_directory)
    perf_summary = PerformanceSummary(input_directory)

    perf_df = perf_summary.to_df()

    if len(perf_df) < 1:
        logger.warning(f"No performance data found in {input_directory}.")
        return  # nothing left to do

    # Add jobapi data about the job to dataframe
    perf_df = add_squid_api_data(perf_df)

    # Set index to include branch configuration/scenario columns
    perf_df, scenario_cols = set_index_scenario_cols(perf_df)
    # preserve copy before updating for stat report
    original_perf_df = perf_df.copy()

    # Write to file
    out_file = output_directory / "log_summary"
    if output_hdf:
        out_file = out_file.with_suffix(".hdf")
        perf_df.to_hdf(out_file, key="worker_data")
    else:
        out_file = out_file.with_suffix(".csv")
        perf_df.to_csv(out_file)

    # Clean up performance logs
    perf_summary.clean_perf_logs()

    if verbose:
        print_stat_report(perf_df, scenario_cols)

    if perf_summary.errors > 0:
        logger.warning(
            f'{perf_summary.errors} log row{"s were" if perf_summary.errors > 1 else " was"} unreadable.'
        )
    logger.info(
        f'Performance summary {"hdf" if output_hdf else "csv"} can be found at {out_file}, with '
        f'{perf_df.shape[0]} row{"s" if perf_df.shape[0] > 1 else ""}.'
    )
    return original_perf_df



"""
====================
psimulate RQ Workers
====================

Launching RQ workers is done by specifying Python import paths to worker implementations.
The actual worker classes and functions are hidden from view here as they should not be
referenced directly in the main `psimulate` process.

There are three pieces to an RQ worker.

#. The Worker Class. The worker class and is a normal Python class that inherits from
   `rq.Worker` (this default implementation is fully functional on its own). The worker
   class governs interaction with the Redis DB where the jobs and results are stored,
   the retry behavior on DB connection errors, and the forking behavior when it runs
   a job, and a number of other things. The `ResilientWorker` defined in the `core` module
   primarily makes some small adjustments to the logging and retry behavior and squashes
   some undesirable forking behavior on clusters that can orphan child processes.
#. The Exception handler. The exception handler is a plain python function that determines
   what happens when an actual job fails due to something other than a Redis DB connection
   issue. These failures usually occur when there is a code error, a data error, or
   an inability to access some resource within the job itself (filesystem, database, etc.).
#. The Work Horse.  The work horse is also a plain python function. It is the actual
   execution logic of a job (a job is a blob of parameters retrieved by the worker class
   and passed in to the work horse). The function takes a single argument which is
   deserialized off the job queue in the connected Redis database. The body of the function
   can really do anything at that point. It can then return a value which will be serialized
   (using pickle, I'm pretty sure) and stored back in the results section of the connected
   Redis DB.

The `build_launch_script` function provided by this package produces a shell file
used to spin up our (slightly customized) general purpose worker class with an exception
handler that will retry jobs a couple of times.

"""
from vivarium_cluster_tools.psimulate import COMMANDS
from vivarium_cluster_tools.psimulate.worker.core import build_launch_script
from vivarium_cluster_tools.psimulate.worker.load_test_work_horse import (
    LOAD_TEST_WORK_HORSE_IMPORT_PATH,
)

# Work horses are specific to a kind of job.
from vivarium_cluster_tools.psimulate.worker.vivarium_work_horse import (
    VIVARIUM_WORK_HORSE_IMPORT_PATH,
)

# All work horses available to psimulate
WORK_HORSE_PATHS = {
    COMMANDS.run: VIVARIUM_WORK_HORSE_IMPORT_PATH,
    COMMANDS.restart: VIVARIUM_WORK_HORSE_IMPORT_PATH,
    COMMANDS.expand: VIVARIUM_WORK_HORSE_IMPORT_PATH,
    COMMANDS.load_test: LOAD_TEST_WORK_HORSE_IMPORT_PATH,
}

# Clean up the namespace
del VIVARIUM_WORK_HORSE_IMPORT_PATH
del LOAD_TEST_WORK_HORSE_IMPORT_PATH
del COMMANDS



"""
=========
psimulate
=========

Parallel runner for :mod:`vivarium` jobs.

"""
from typing import NamedTuple


class __Commands(NamedTuple):
    run: str
    restart: str
    expand: str
    load_test: str


COMMANDS = __Commands(*__Commands._fields)

del NamedTuple
del __Commands



"""
=============================
Redis Database Initialization
=============================

Creates redis databases to store job parameters and results.

"""

import atexit
import math
import socket
import subprocess
import sys
import time
from pathlib import Path
from typing import TextIO

from loguru import logger

from vivarium_cluster_tools.psimulate.environment import ENV_VARIABLES

DEFAULT_NUM_REDIS_DBS = -1
DEFAULT_WORKERS_PER_REDIS_INSTANCE = 1000


def launch_redis_processes(
    num_processes: int,
    num_workers: int,
    redis_logging_root: Path,
) -> list[tuple[str, int]]:
    num_processes = _get_num_redis_dbs(num_processes, num_workers)

    hostname = ENV_VARIABLES.HOSTNAME.value
    redis_ports = []
    for i in range(num_processes):
        port = _get_random_free_port()
        logger.info(f"Starting Redis Broker at {hostname}:{port}")
        redis_log = (redis_logging_root / f"redis.p{port}.log").open("a")
        _launch_redis(port, stdout=redis_log, stderr=redis_log)
        redis_ports.append((hostname, port))
    time.sleep(5)  # Give the dbs a few seconds to spin up.
    return redis_ports


def _get_num_redis_dbs(num_processes: int, num_workers: int) -> int:
    if num_processes == DEFAULT_NUM_REDIS_DBS:
        num_processes = int(math.ceil(num_workers / DEFAULT_WORKERS_PER_REDIS_INSTANCE))
    else:
        min_workers = _expected_sufficient_workers(num_processes)
        if num_workers < min_workers:
            logger.warning(
                f"With {num_processes} queues, you should have >> {min_workers} workers, "
                "but you only have {num_workers}. "
                "Failure to allocate sufficent workers may result in jobs not being processed. "
                "Consider increasing the number of workers, or decreasing the number of redis queues."
            )
    return num_processes


def _launch_redis(
    port: int, stdout: TextIO = sys.stdout, stderr: TextIO = sys.stderr
) -> subprocess.Popen:
    stdout.write(f">>>>>>>> Starting log for redis-server on port {port}\n")
    stdout.flush()
    try:
        # inline config for redis server.
        redis_process = subprocess.Popen(
            [
                "redis-server",
                "--port",
                f"{port}",
                "--timeout",
                "2",
                "--loglevel",
                "debug",
                "--protected-mode",
                "no",
            ],
            stdout=stdout,
            stderr=stderr,
        )
    except FileNotFoundError:
        raise OSError(
            "In order for redis to launch you need both the redis client and the python bindings. "
            "You seem to be missing the redis client.  Do 'conda install redis' and try again. If "
            "failures continue you may need to download redis yourself, make it and add it to PATH."
        )
    atexit.register(redis_process.kill)
    return redis_process


def _get_random_free_port() -> int:
    """Get a random free port number.

    Notes
    -----
    This implementation is vulnerable to rare race conditions where some other
    process gets the same port after we free our socket but before we use the port
    number we got. Should be so rare in practice that it doesn't matter.
    """
    s = socket.socket()
    s.bind(("", 0))
    port = s.getsockname()[1]
    s.close()
    return port


def _expected_sufficient_workers(num_queues) -> int:
    """Estimate the number of workers needed.

    The intent is to provide a rough estimate of the number of workers needed to
    ensure that each queue gets at least one worker. cf. https://w.wiki/7bnb
    """
    return int(math.ceil(num_queues * (math.log(num_queues) + 0.57)))



"""
==============================
Results Processing CLI Options
==============================

Command line options for configuring results handling in psimulate runs.

"""
import click

from vivarium_cluster_tools.cli_tools import MINUTES_OR_NONE

with_no_batch = click.option(
    "--no-batch", is_flag=True, help="Don't batch results, write them as they come in."
)
with_no_cleanup = click.option(
    "--no-cleanup",
    is_flag=True,
    hidden=True,
    help="Hidden developer option, if flagged, don't automatically "
    "cleanup results directory on failure.",
)
backup_freq = click.option(
    "--backup-freq",
    type=MINUTES_OR_NONE,
    default="30",
    show_default=True,
    help="Interval in minutes between saving backups. Set to 'None' or 'none' to disable backups.",
)
